import os

from django.conf import settings
from django.core.management.base import BaseCommand

from ... import settings as default_settings
from ...choices import generate_ts


class Command(BaseCommand):
    help = 'Creates a static choices.d.ts file for django-js-choices'

    def get_location(self):
        output_path = getattr(settings, 'DJANGO_TS_BRIDGE_CHOICE_DEFINITION_OUTPUT_PATH', default_settings.DJANGO_TS_BRIDGE_CHOICE_DEFINITION_OUTPUT_PATH)
        return os.path.join(settings.BASE_DIR, output_path)

    def handle(self, *args, **options):
        location = self.get_location()
        content = generate_ts()
        with open(location, 'w') as type_file:
            type_file.write(content)
        self.stdout.write(self.style.SUCCESS(f'Choices definition file sucessfully generated at {location}'))



from django.apps import apps
from django.conf import settings
from django.template import loader

from . import settings as default_settings
from .core import is_model_choice

def generate_choices():
    choices = {}
    for app_config in apps.get_app_configs():
        for model in app_config.get_models():
            choice_enums = [cls_attribute for cls_attribute in model.__dict__.values() if is_model_choice(cls_attribute)]
            if len(choice_enums):
                choice_definitions = {}
                for enum in choice_enums:
                    choice_definition = {}
                    for name in enum.names:
                        choice_definition[name] = enum[name]
                    choice_definitions[enum.__name__] = choice_definition
                choices[app_config.label] = {}
                choices[app_config.label][model._meta.object_name] = choice_definitions
    return choices

def generate_ts():
    choices = generate_choices()
    ts_var_name = getattr(settings, 'DJANGO_TS_BRIDGE_CHOICE_VAR_NAME', default_settings.DJANGO_TS_BRIDGE_CHOICE_VAR_NAME)
    ts_content = loader.render_to_string(
        'django_ts_bridge/choices_ts.tpl',
        {
            'choices': choices,
            'ts_var_name': ts_var_name,
        },
    )
    return ts_content



import importlib
import inspect
import os
from types import ModuleType
from django.apps import AppConfig
from django.conf import settings
from django.core.management.base import BaseCommand
from rest_framework.serializers import Serializer

from ... import model_defs, settings as default_settings
from ...fields import ts_interface, get_ts

class Command(BaseCommand):
    help = 'Creates a TypeScript definition file for all models registered in DRF serializers under DJANGO_TS_BRIDGE_MODEL_DEFINITION_OUTPUT_PATH'

    def get_location(self):
        output_path = getattr(settings, 'DJANGO_TS_BRIDGE_MODEL_DEFINITION_OUTPUT_PATH', default_settings.DJANGO_TS_BRIDGE_MODEL_DEFINITION_OUTPUT_PATH)
        return os.path.join(settings.BASE_DIR, output_path)

    def handle(self, *args, **options):
        self.excluded_apps = getattr(settings, 'DJANGO_TS_BRIDGE_EXCLUDED_APPS', default_settings.DJANGO_TS_BRIDGE_EXCLUDED_APPS)
        from django.apps import apps
        for app_config in apps.get_app_configs():
            if app_config.name not in self.excluded_apps:
                self.handle_app_config(app_config)
        # Remove Serializer from type name since it's redundant in TS
        choices_import_path = getattr(settings, 'DJANGO_TS_BRIDGE_CHOICE_DEFINITION_IMPORT_PATH', default_settings.DJANGO_TS_BRIDGE_CHOICE_DEFINITION_IMPORT_PATH)
        ts = get_ts().replace('Serializer', '')
        ts = f'import {{ Choices }} from \'{choices_import_path}\';\n' + ts
        type_file_location = self.get_location()
        with open(type_file_location, 'w') as type_file:
            type_file.write(ts)
        self.stdout.write(self.style.SUCCESS(f'Type file sucessfully generated at {type_file_location}'))

    def handle_app_config(self, app_config: AppConfig):
        try: #Check to see the app has serializers
            serializers_module: ModuleType = importlib.import_module(app_config.name + '.serializers')
        except ImportError:
            return
        mappings = model_defs.create_model_mappings(app_config)
        serializers = inspect.getmembers(serializers_module, lambda member: self.is_serializer(member, serializers_module))
        for name, serializer in  serializers:
            # Get the class def and apply the ts_interface decorator to it
            base_class = getattr(serializers_module, name)
            ts_interface(mapping_overrides=mappings)(base_class)
    
    def is_serializer(self, member: object, module):
        """Checks to see if the given member is a serializer class and is a part of the given module"""
        return inspect.isclass(member) and issubclass(member, Serializer) and member.__module__ == module.__name__



import logging
from rest_framework import serializers
from .mappings import mappings

_LOG = logging.getLogger(f"django-typomatic.{__name__}")

# Serializers
__serializers = dict()
# Custom serializers.Field to TS Type mappings
__field_mappings = dict()
# Custom field_name to TS Type overrides
__mapping_overrides = dict()


def ts_field(ts_type: str, context='default'):
    '''
    Any valid Django Rest Framework Serializer Field with this class decorator will
    be added to a list in a __field_mappings dictionary.
    Useful to define the type mapping of custom serializer Fields.
    e.g.
    @ts_field('string')
    class CustomFieldField(serializers.Field):
        def to_internal_value(self, data):
            pass
        def to_representation(self, obj):
            pass
    '''
    def decorator(cls):
        if issubclass(cls, serializers.Field):
            if context not in __field_mappings:
                __field_mappings[context] = dict()
            if cls not in __field_mappings[context]:
                __field_mappings[context][cls] = ts_type
        return cls
    return decorator


def ts_interface(context='default', mapping_overrides=None):
    '''
    Any valid Django Rest Framework Serializers with this class decorator will
    be added to a list in a dictionary.
    Optional parameters:
    'context': Will create separate dictionary keys per context.
        Otherwise, all values will be inserted into a list with a key of 'default'.
    'mapping_overrides': Dictionary of field_names to TS types
        Useful to properly serialize ModelSerializer runtime properties and ReadOnlyFields.
    e.g.
    @ts_interface(context='internal', mapping_overrides={"baz" : "string[]"})
    class Foo(serializers.Serializer):
        bar = serializer.IntegerField()
        baz = serializer.ReadOnlyField(source='baz_property')
    '''
    def decorator(cls):
        if issubclass(cls, serializers.Serializer):
            if context not in __field_mappings:
                __field_mappings[context] = dict()
            if context not in __serializers:
                __serializers[context] = []
            __serializers[context].append(cls)
            if mapping_overrides:
                if context not in __mapping_overrides:
                    __mapping_overrides[context] = dict()
                if cls not in __mapping_overrides[context]:
                    __mapping_overrides[context][cls] = mapping_overrides
        return cls
    return decorator


def __process_field(field_name, field, context, serializer):
    '''
    Generates and returns a tuple representing the Typescript field name and Type.
    '''
    if hasattr(field, 'child'):
        is_many = True
        field_type = type(field.child)
    elif hasattr(field, 'child_relation'):
        is_many = True
        field_type = type(field.child_relation)
    else:
        is_many = False
        field_type = type(field)

    if field_type in __serializers[context]:
        ts_type = field_type.__name__
    elif field_type in __field_mappings[context]:
        ts_type = __field_mappings[context].get(field_type, 'any')
    elif (context in __mapping_overrides) and (serializer in __mapping_overrides[context]) and field_name in __mapping_overrides[context][serializer]:
        ts_type = __mapping_overrides[context][serializer].get(
            field_name, 'any')
    else:
        ts_type = mappings.get(field_type, 'any')
    if is_many:
        # Make sure all the types that are unioned are converted to be an array
        ts_types = ts_type.split('|')
        ts_type = '|'.join([sub_type + '[]' for sub_type in ts_types])
    return (field_name, ts_type)


def __get_ts_interface(serializer, context):
    '''
    Generates and returns a Typescript Interface by iterating
    through the serializer fields of the DRF Serializer class
    passed in as a parameter, and mapping them to the appropriate Typescript
    data type.
    '''
    name = serializer.__name__
    _LOG.debug(f"Creating interface for {name}")
    fields = []
    if hasattr(serializer, 'get_fields'):
        instance = serializer()
        fields = instance.get_fields().items()
    else:
        fields = serializer._declared_fields.items()
    ts_fields = []
    for key, value in fields:
        ts_field = __process_field(key, value, context, serializer)
        if value.read_only or not value.required:
            op = '?:'
        else:
            op = ':'
        ts_fields.append(f"    {ts_field[0]}{op} {ts_field[1]};")
    collapsed_fields = '\n'.join(ts_fields)
    return f'export interface {name} {{\n{collapsed_fields}\n}}\n\n'


def __generate_interfaces(context):
    if context not in __serializers:
        return []
    return [__get_ts_interface(serializer, context)
            for serializer in __serializers[context]]


def generate_ts(output_path, context='default'):
    '''
    When this function is called, a Typescript interface will be generated
    for each DRF Serializer in the serializers dictionary, depending on the
    optional context parameter provided. If the parameter is ignored, all
    serializers in the default value, 'default' will be iterated over and a
    list of Typescript interfaces will be returned via a list comprehension.

    The Typescript interfaces will then be outputted to the file provided.
    '''
    with open(output_path, 'w') as output_file:
        interfaces = __generate_interfaces(context)
        print(interfaces)
        output_file.write(''.join(interfaces))


def get_ts(context='default'):
    '''
    Similar to generate_ts. But rather than outputting the generated 
    interfaces to the specified file, will return the generated interfaces 
    as a raw string.
    '''
    interfaces = __generate_interfaces(context)
    return ''.join(interfaces)



from typing import List
from django.apps import AppConfig
from django.db import models
from .core import is_model_choice


def create_enum(app_config, model, field):
    enums: List[models.Choices] = [cls_attribute for cls_attribute in model.__dict__.values() if is_model_choice(cls_attribute)]
    enum_name = ''
    ts_type = ''
    # Try and find the enum definition of the choices field. TODO check to see if this will ever bite me in the ass.
    # cause 2 enums could in theory have the same definition, although that would be dumb and highly unlikely
    for enum in enums:
        if enum.choices == field.choices:
            enum_name = enum.__name__
            break
    # No enum was defined so set the type to be a list of all possible raw values
    if enum_name == '':
        for choice in field.choices:
            value = choice[0]
            if isinstance(value, int):
                ts_type += f'{choice[0]} | '
            else:
                ts_type += f'\'{choice[0]}\' | '
        ts_type = ts_type[:-2]
    else:
        ts_type = f'typeof Choices.{app_config.label}.{model._meta.object_name}.{enum_name}'
    return ts_type

def create_model_mappings(app_config: AppConfig):
    """Updates the custom mappings with related models and their type in ts"""
    mappings = {}
    for model in app_config.get_models():
        for field in model._meta.get_fields():
            field_name = field.name
            if field.related_model is not None:
                ts_type = f'{field.related_model.__name__}|number' # Always allow the related models to be defined by their pks
                mappings[field_name] = ts_type
            elif field.choices is not None:
                mappings[field_name] = create_enum(app_config, model, field)
    return mappings


from django import template
from django.utils.safestring import mark_safe

register = template.Library()

@register.filter_function
def js_literal(value):
    js_literal_value = 'null'
    if value is not None:
        if isinstance(value, str):
            js_literal_value = f'"{value}"'
        else:
            js_literal_value = value
    return js_literal_value



DJANGO_TS_BRIDGE_EXCLUDED_APPS = ['rest_framework', 'wagtail.api.v2']
DJANGO_TS_BRIDGE_CHOICE_DEFINITION_IMPORT_PATH = 'ts/django/django-model-choices'
DJANGO_TS_BRIDGE_CHOICE_DEFINITION_OUTPUT_PATH = 'src/ts/django/django-model-choices.ts'
DJANGO_TS_BRIDGE_MODEL_DEFINITION_OUTPUT_PATH = 'src/ts/@types/django-models.d.ts'
DJANGO_TS_BRIDGE_URL_OUTPUT_PATH = 'src/django/urls.json'
DJANGO_TS_BRIDGE_CHOICE_VAR_NAME = 'Choices'
DJANGO_TS_BRIDGE_URLS_VAR_NAME = 'Urls'


from django.core.management import call_command
from django.core.management.base import BaseCommand

class Command(BaseCommand):
    help = 'Creates all TS files used to define the structure of Django models'

    def handle(self, *args, **options):
        call_command('create_ts_choice_defs', *args, **options)
        call_command('create_ts_serializer_types', *args, **options)
        call_command('create_ts_urls', *args, **options)


import setuptools

with open('README.md') as readme:
    long_description = readme.read()

with open('requirements.txt') as requirements_file:
    requirements = requirements_file.readlines()

setuptools.setup(
    name='django-ts-bridge',
    version='1.3.1',
    author='ChickenF622',
    author_email='chickenf622@gmail.com',
    description='Management commands in Django that give defintions for Django models returned by DRF along with any model fields limited by choices',
    long_description=long_description,
    include_package_data=True,
    package_data= {
        'django_ts_bridge': [
            'templates/django_ts_bridge/*.tpl'
        ]
    },
    install_requires=requirements,
    url='https://gitlab.com/ChickenF622/django-ts-bridge',
    #package_dir={'': 'django_ts_bridge'},
    packages=setuptools.find_packages(exclude=['tests']),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
    python_requires=">=3.8"
)



import json
import os
from django.conf import settings
from django.contrib.admindocs.views import simplify_regex
from django.core.exceptions import ImproperlyConfigured, ViewDoesNotExist
from django.core.management.base import BaseCommand
from django.template import loader
from django.urls.resolvers import URLPattern, URLResolver

from ... import settings as default_settings

def describe_pattern(p):
    return str(p.pattern)

class Command(BaseCommand):
    def handle(self, *args, **options):
        views = self.extract_views_from_urlpatterns()
        urls = {}
        for (_, regex, url_name) in views:
            urls[url_name] = simplify_regex(regex)
        included_urls = {}
        if 'DJANGO_TS_BRIDGE_INCLUDE_URL_NAMESPACES' not in settings:
            raise ImproperlyConfigured('DJANGO_TS_BRIDGE_INCLUDE_URL_NAMESPACES must be set to generate URLs')
        namespaces = tuple(f'{namespace}:' for namespace in settings.DJANGO_TS_BRIDGE_INCLUDE_URL_NAMESPACES)
        for url_name, url in urls.items():
            if url_name is not None and url_name.startswith(namespaces):
                included_urls[url_name] = url
        url_file_location = self.get_location()
        with open(url_file_location, 'w') as url_file:
            url_file.write(json.dumps(included_urls))
        self.stdout.write(self.style.SUCCESS(f'URL file sucessfully generated at {url_file_location}'))


    def get_location(self):
        output_path = getattr(settings, 'DJANGO_TS_BRIDGE_URL_OUTPUT_PATH', default_settings.DJANGO_TS_BRIDGE_URL_OUTPUT_PATH)
        return os.path.join(settings.BASE_DIR, output_path)

    def extract_views_from_urlpatterns(self, urlpatterns=None, base='', namespace=None):
        """
        Return a list of views from a list of urlpatterns.
        Each object in the returned list is a three-tuple: (view_func, regex, name)
        """
        if urlpatterns is None:
            urlconf = __import__(settings.ROOT_URLCONF, {}, {}, [''])
            urlpatterns = urlconf.urlpatterns
        views = []
        for p in urlpatterns:
            if isinstance(p, URLPattern):
                try:
                    if not p.name:
                        name = p.name
                    elif namespace:
                        name = '{0}:{1}'.format(namespace, p.name)
                    else:
                        name = p.name
                    pattern = describe_pattern(p)
                    views.append((p.callback, base + pattern, name))
                except ViewDoesNotExist:
                    continue
            elif isinstance(p, URLResolver):
                try:
                    patterns = p.url_patterns
                except ImportError:
                    continue
                if namespace and p.namespace:
                    _namespace = '{0}:{1}'.format(namespace, p.namespace)
                else:
                    _namespace = (p.namespace or namespace)
                pattern = describe_pattern(p)
                views.extend(self.extract_views_from_urlpatterns(patterns, base + pattern, namespace=_namespace))
            elif hasattr(p, '_get_callback'):
                try:
                    views.append((p._get_callback(), base + describe_pattern(p), p.name))
                except ViewDoesNotExist:
                    continue
            elif hasattr(p, 'url_patterns') or hasattr(p, '_get_url_patterns'):
                try:
                    patterns = p.url_patterns
                except ImportError:
                    continue
                views.extend(self.extract_views_from_urlpatterns(patterns, base + describe_pattern(p), namespace=namespace))
            else:
                raise TypeError("%s does not appear to be a urlpattern object" % p)
        return views
